@@ -49,6 +49,10 @@
  * 11/02/2004 - Doxygenified the source, modularized it a bit. Needs a 
  * 	lot more work, but this is a start. Wouter Verhelst
  * 	<wouter@debian.org>
+ * 16/03/2010 - Add IPv6 support.
+ *	Kitt Tientanopajai <kitt@kitty.in.th>
+ *	Neutron Soutmun <neo.neutron@gmail.com>
+ *	Suriya Soutmun <darksolar@gmail.com>
  */
 
 /* Includes LFS defines, which defines behaviours of some of the following
@@ -70,8 +74,8 @@
 #include <signal.h>		/* sigaction */
 #include <errno.h>
 #include <netinet/tcp.h>
-#include <netinet/in.h>		/* sockaddr_in, htons, in_addr */
-#include <netdb.h>		/* hostent, gethostby*, getservby* */
+#include <netinet/in.h>
+#include <netdb.h>
 #include <syslog.h>
 #include <unistd.h>
 #include <stdio.h>
@@ -149,6 +153,7 @@ gchar* rungroup=NULL;
 #define F_AUTOREADONLY 8  /**< flag to tell us a file is set to autoreadonly */
 #define F_SPARSE 16	  /**< flag to tell us copyronwrite should use a sparse file */
 #define F_SDP 32	  /**< flag to tell us the export should be done using the Socket Direct Protocol for RDMA */
+#define F_SYNC 64	  /**< Whether to fsync() after a write */
 GHashTable *children;
 char pidfname[256]; /**< name of our PID file */
 char pidftemplate[256]; /**< template to be used for the filename of the PID file */
@@ -177,6 +182,7 @@ typedef struct {
 	char* authname;      /**< filename of the authorization file */
 	int flags;           /**< flags associated with this exported file */
 	int socket;	     /**< The socket of this server. */
+	int socket_family;   /**< family of the socket */
 	VIRT_STYLE virtstyle;/**< The style of virtualization, if any */
 	uint8_t cidrlen;     /**< The length of the mask when we use
 				  CIDR-style virtualization */
@@ -302,10 +308,14 @@ inline void readit(int f, void *buf, size_t len) {
 	ssize_t res;
 	while (len > 0) {
 		DEBUG("*");
-		if ((res = read(f, buf, len)) <= 0)
-			err("Read failed: %m");
-		len -= res;
-		buf += res;
+		if ((res = read(f, buf, len)) <= 0) {
+			if(errno != EAGAIN) {
+				err("Read failed: %m");
+			}
+		} else {
+			len -= res;
+			buf += res;
+		}
 	}
 }
 
@@ -333,13 +343,12 @@ inline void writeit(int f, void *buf, size_t len) {
  */
 void usage() {
 	printf("This is nbd-server version " VERSION "\n");
-	printf("Usage: [ip:]port file_to_export [size][kKmM] [-l authorize_file] [-r] [-m] [-c] [-C configuration file] [-p PID file name] [-o section name]\n"
+	printf("Usage: [ip:|ip6@]port file_to_export [size][kKmM] [-l authorize_file] [-r] [-m] [-c] [-C configuration file] [-p PID file name] [-o section name]\n"
 	       "\t-r|--read-only\t\tread only\n"
 	       "\t-m|--multi-file\t\tmultiple file\n"
 	       "\t-c|--copy-on-write\tcopy on write\n"
 	       "\t-C|--config-file\tspecify an alternate configuration file\n"
 	       "\t-l|--authorize-file\tfile with list of hosts that are allowed to\n\t\t\t\tconnect.\n"
-	       "\t-a|--idle-time\t\tmaximum idle seconds; server terminates when\n\t\t\t\tidle time exceeded\n"
 	       "\t-p|--pid-file\t\tspecify a filename to write our PID to\n"
 	       "\t-o|--output-config\toutput a config file section for what you\n\t\t\t\tspecified on the command line, with the\n\t\t\t\tspecified section name\n\n"
 	       "\tif port is set to 0, stdin is used (for running from inetd)\n"
@@ -388,7 +397,6 @@ SERVER* cmdline(int argc, char *argv[]) {
 		{"multi-file", no_argument, NULL, 'm'},
 		{"copy-on-write", no_argument, NULL, 'c'},
 		{"authorize-file", required_argument, NULL, 'l'},
-		{"idle-time", required_argument, NULL, 'a'},
 		{"config-file", required_argument, NULL, 'C'},
 		{"pid-file", required_argument, NULL, 'p'},
 		{"output-config", required_argument, NULL, 'o'},
@@ -408,18 +416,30 @@ SERVER* cmdline(int argc, char *argv[]) {
 	serve=g_new0(SERVER, 1);
 	serve->authname = g_strdup(default_authname);
 	serve->virtstyle=VIRT_IPLIT;
-	while((c=getopt_long(argc, argv, "-a:C:cl:mo:rp:", long_options, &i))>=0) {
+	while((c=getopt_long(argc, argv, "-C:cl:mo:rp:", long_options, &i))>=0) {
 		switch (c) {
 		case 1:
 			/* non-option argument */
 			switch(nonspecial++) {
 			case 0:
-				addr_port=g_strsplit(optarg, ":", 2);
+				if(strchr(optarg, ':') == strrchr(optarg, ':')) {
+					addr_port=g_strsplit(optarg, ":", 2);
+
+					/* Check for "@" - maybe user using this separator
+						 for IPv4 address */
+					if(!addr_port[1]) {
+						g_strfreev(addr_port);
+						addr_port=g_strsplit(optarg, "@", 2);
+					}
+				} else {
+					addr_port=g_strsplit(optarg, "@", 2);
+				}
+
 				if(addr_port[1]) {
 					serve->port=strtol(addr_port[1], NULL, 0);
 					serve->listenaddr=g_strdup(addr_port[0]);
 				} else {
-					serve->listenaddr=g_strdup("0.0.0.0");
+					serve->listenaddr=NULL;
 					serve->port=strtol(addr_port[0], NULL, 0);
 				}
 				g_strfreev(addr_port);
@@ -522,6 +542,112 @@ void remove_server(gpointer s) {
 	g_free(server);
 }
 
+/**
+ * duplicate server
+ * @param s the old server we want to duplicate
+ * @return new duplicated server
+ **/
+SERVER* dup_serve(SERVER *s) {
+	SERVER *serve = NULL;
+
+	serve=g_new0(SERVER, 1);
+	if (serve == NULL)
+		return NULL;
+
+	if (s->exportname)
+		serve->exportname = g_strdup(s->exportname);
+
+	serve->expected_size = s->expected_size;
+
+	if (s->listenaddr)
+		serve->listenaddr = g_strdup(s->listenaddr);
+
+	serve->port = s->port;
+
+	if (s->authname)
+		serve->authname = strdup(s->authname);
+
+	serve->flags = s->flags;
+	serve->socket = serve->socket;
+	serve->socket_family = serve->socket_family;
+	serve->cidrlen = s->cidrlen;
+
+	if (s->prerun)
+		serve->prerun = g_strdup(s->prerun);
+
+	if (s->postrun)
+		serve->postrun = g_strdup(s->postrun);
+
+	return serve;
+}
+
+/**
+ * append new server to array
+ * @param s server
+ * @param a server array
+ * @return 0 success, -1 error
+ */
+int append_serve(SERVER *s, GArray *a)
+{
+	SERVER *ns = NULL;
+	struct addrinfo hints;
+	struct addrinfo *ai = NULL;
+	struct addrinfo *rp = NULL;
+	char   host[NI_MAXHOST];
+	gchar  *port = NULL;
+	int e;
+	int ret;
+
+	if(!s) {
+		err("Invalid parsing server");
+		return -1;
+	}
+
+	port = g_strdup_printf("%d", s->port);
+
+	memset(&hints,'\0',sizeof(hints));
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_flags = AI_ADDRCONFIG | AI_PASSIVE;
+	hints.ai_protocol = IPPROTO_TCP;
+
+	e = getaddrinfo(s->listenaddr, port, &hints, &ai);
+
+	if (port)
+		g_free(port);
+
+	if(e == 0) {
+		for (rp = ai; rp != NULL; rp = rp->ai_next) {
+			e = getnameinfo(rp->ai_addr, rp->ai_addrlen, host, sizeof(host), NULL, 0, NI_NUMERICHOST);
+
+			if (e != 0) { // error
+				fprintf(stderr, "getnameinfo: %s\n", gai_strerror(e));
+				continue;
+			}
+
+			// duplicate server and set listenaddr to resolved IP address
+			ns = dup_serve (s);
+			if (ns) {
+				ns->listenaddr = g_strdup(host);
+				ns->socket_family = rp->ai_family;
+				g_array_append_val(a, *ns);
+				free(ns);
+				ns = NULL;
+			}
+		}
+
+		ret = 0;
+	} else {
+		fprintf(stderr, "getaddrinfo failed on listen host/address: %s (%s)\n", s->listenaddr ? s->listenaddr : "any", gai_strerror(e));
+		ret = -1;
+	}
+
+	if (ai)
+		freeaddrinfo(ai);
+
+	return ret;
+}
+
 /**
  * Parse the config file.
  *
@@ -550,6 +676,7 @@ GArray* parse_cfile(gchar* f, GError** e) {
 		{ "copyonwrite", FALSE,	PARAM_BOOL,	NULL, F_COPYONWRITE },
 		{ "sparse_cow",	FALSE,	PARAM_BOOL,	NULL, F_SPARSE },
 		{ "sdp",	FALSE,	PARAM_BOOL,	NULL, F_SDP },
+		{ "sync",	FALSE,  PARAM_BOOL,	NULL, F_SYNC },
 		{ "listenaddr", FALSE,  PARAM_STRING,   NULL, 0 },
 	};
 	const int lp_size=sizeof(lp)/sizeof(PARAM);
@@ -575,7 +702,7 @@ GArray* parse_cfile(gchar* f, GError** e) {
 	retval = g_array_new(FALSE, TRUE, sizeof(SERVER));
 	if(!g_key_file_load_from_file(cfile, f, G_KEY_FILE_KEEP_COMMENTS |
 			G_KEY_FILE_KEEP_TRANSLATIONS, &err)) {
-		g_set_error(e, errdomain, CFILE_NOTFOUND, "Could not open config file.");
+		g_set_error(e, errdomain, CFILE_NOTFOUND, "Could not open config file %s.", f);
 		g_key_file_free(cfile);
 		return retval;
 	}
@@ -596,8 +723,9 @@ GArray* parse_cfile(gchar* f, GError** e) {
 		lp[5].target=&(s.prerun);
 		lp[6].target=&(s.postrun);
 		lp[7].target=lp[8].target=lp[9].target=
-				lp[10].target=lp[11].target=&(s.flags);
-		lp[12].target=&(s.listenaddr);
+				lp[10].target=lp[11].target=
+				lp[12].target=&(s.flags);
+		lp[13].target=&(s.listenaddr);
 
 		/* After the [generic] group, start parsing exports */
 		if(i==1) {
@@ -683,10 +811,9 @@ GArray* parse_cfile(gchar* f, GError** e) {
 		virtstyle=NULL;
 		/* Don't append values for the [generic] group */
 		if(i>0) {
-			if(!s.listenaddr) {
-				s.listenaddr = g_strdup("0.0.0.0");
-			}
-			g_array_append_val(retval, s);
+			s.socket_family = AF_UNSPEC;
+
+			append_serve(&s, retval);
 		}
 #ifndef WITH_SDP
 		if(s.flags & F_SDP) {
@@ -883,6 +1010,7 @@ ssize_t rawexpwrite(off_t a, char *buf, size_t len, CLIENT *client) {
 	int fhandle;
 	off_t foffset;
 	size_t maxbytes;
+	ssize_t retval;
 
 	if(get_filepos(client->export, a, &fhandle, &foffset, &maxbytes))
 		return -1;
@@ -892,7 +1020,11 @@ ssize_t rawexpwrite(off_t a, char *buf, size_t len, CLIENT *client) {
 	DEBUG4("(WRITE to fd %d offset %llu len %u), ", fhandle, foffset, len);
 
 	myseek(fhandle, foffset);
-	return write(fhandle, buf, len);
+	retval = write(fhandle, buf, len);
+	if(client->server->flags & F_SYNC) {
+		fsync(fhandle);
+	}
+	return retval;
 }
 
 /**
@@ -1200,6 +1332,7 @@ void setupexport(CLIENT* client) {
 	for(i=0; ; i++) {
 		FILE_INFO fi;
 		gchar *tmpname;
+		gchar* error_string;
 		mode_t mode = (client->server->flags & F_READONLY) ? O_RDONLY : O_RDWR;
 
 		if(multifile) {
@@ -1224,7 +1357,10 @@ void setupexport(CLIENT* client) {
 		if(fi.fhandle == -1) {
 			if(multifile && i>0)
 				break;
-			err("Could not open exported file: %m");
+			error_string=g_strdup_printf(
+				"Could not open exported file %s: %%m",
+				tmpname);
+			err(error_string);
 		}
 		fi.startoff = laststartoff + lastsize;
 		g_array_append_val(client->export, fi);
@@ -1330,17 +1466,36 @@ void serveconnection(CLIENT *client) {
  * stored in client->clientname.
  **/
 void set_peername(int net, CLIENT *client) {
-	struct sockaddr_in addrin;
-	struct sockaddr_in netaddr;
+	struct sockaddr_storage addrin;
+	struct sockaddr_storage netaddr;
+	struct sockaddr_in  *netaddr4 = NULL;
+	struct sockaddr_in6 *netaddr6 = NULL;
 	size_t addrinlen = sizeof( addrin );
-	char *peername;
-	char *netname;
-	char *tmp;
+	struct addrinfo hints;
+	struct addrinfo *ai = NULL;
+	char peername[NI_MAXHOST];
+	char netname[NI_MAXHOST];
+	char *tmp = NULL;
 	int i;
+	int e;
+	int shift;
 
 	if (getpeername(net, (struct sockaddr *) &addrin, (socklen_t *)&addrinlen) < 0)
 		err("getsockname failed: %m");
-	peername = g_strdup(inet_ntoa(addrin.sin_addr));
+
+	getnameinfo((struct sockaddr *)&addrin, (socklen_t)addrinlen,
+		peername, sizeof (peername), NULL, 0, NI_NUMERICHOST);
+
+	memset(&hints, '\0', sizeof (hints));
+	hints.ai_flags = AI_ADDRCONFIG;
+	e = getaddrinfo(peername, NULL, &hints, &ai);
+
+	if(e != 0) {
+		fprintf(stderr, "getaddrinfo failed: %s\n", gai_strerror(e));
+		freeaddrinfo(ai);
+		return;
+	}
+
 	switch(client->server->virtstyle) {
 		case VIRT_NONE:
 			client->exportname=g_strdup(client->server->exportname);
@@ -1356,18 +1511,42 @@ void set_peername(int net, CLIENT *client) {
 			break;
 		case VIRT_CIDR:
 			memcpy(&netaddr, &addrin, addrinlen);
-			netaddr.sin_addr.s_addr>>=32-(client->server->cidrlen);
-			netaddr.sin_addr.s_addr<<=32-(client->server->cidrlen);
-			netname = inet_ntoa(netaddr.sin_addr);
-			tmp=g_strdup_printf("%s/%s", netname, peername);
-			client->exportname=g_strdup_printf(client->server->exportname, tmp);
+			if(ai->ai_family == AF_INET) {
+				netaddr4 = (struct sockaddr_in *)&netaddr;
+				(netaddr4->sin_addr).s_addr>>=32-(client->server->cidrlen);
+				(netaddr4->sin_addr).s_addr<<=32-(client->server->cidrlen);
+
+				getnameinfo((struct sockaddr *) netaddr4, (socklen_t) addrinlen,
+							netname, sizeof (netname), NULL, 0, NI_NUMERICHOST);
+				tmp=g_strdup_printf("%s/%s", netname, peername);
+			}else if(ai->ai_family == AF_INET6) {
+				netaddr6 = (struct sockaddr_in6 *)&netaddr;
+
+				shift = 128-(client->server->cidrlen);
+				i = 3;
+				while(shift >= 32) {
+					((netaddr6->sin6_addr).s6_addr32[i])=0;
+					shift-=32;
+					i--;
+				}
+				(netaddr6->sin6_addr).s6_addr32[i]>>=shift;
+				(netaddr6->sin6_addr).s6_addr32[i]<<=shift;
+
+				getnameinfo((struct sockaddr *)netaddr6, (socklen_t)addrinlen,
+					    netname, sizeof(netname), NULL, 0, NI_NUMERICHOST);
+				tmp=g_strdup_printf("%s/%s", netname, peername);
+			}
+
+			if(tmp != NULL)
+			  client->exportname=g_strdup_printf(client->server->exportname, tmp);
+
 			break;
 	}
 
+	freeaddrinfo(ai);
 	msg4(LOG_INFO, "connect from %s, assigned file is %s", 
 	     peername, client->exportname);
 	client->clientname=g_strdup(peername);
-	g_free(peername);
 }
 
 /**
@@ -1382,7 +1561,7 @@ void destroy_pid_t(gpointer data) {
  * Loop through the available servers, and serve them. Never returns.
  **/
 int serveloop(GArray* servers) {
-	struct sockaddr_in addrin;
+	struct sockaddr_storage addrin;
 	socklen_t addrinlen=sizeof(addrin);
 	SERVER *serve;
 	int i;
@@ -1478,24 +1657,52 @@ int serveloop(GArray* servers) {
  * @param serve the server we want to connect.
  **/
 void setup_serve(SERVER *serve) {
-	struct sockaddr_in addrin;
+	struct sockaddr_storage addrin;
+	struct addrinfo hints;
+	struct addrinfo *ai = NULL;
 	struct sigaction sa;
 	int addrinlen = sizeof(addrin);
 	int sock_flags;
-	int af;
 #ifndef sun
 	int yes=1;
 #else
 	char yes='1';
 #endif /* sun */
+	gchar *port = NULL;
+	int e;
+
+	memset(&hints,'\0',sizeof(hints));
+	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG | AI_NUMERICSERV;
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_family = serve->socket_family;
+
+	port = g_strdup_printf ("%d", serve->port);
+	if (port == NULL)
+		return;
+
+	e = getaddrinfo(serve->listenaddr,port,&hints,&ai);
+
+	g_free(port);
+
+	if(e != 0) {
+		fprintf(stderr, "getaddrinfo failed: %s\n", gai_strerror(e));
+		serve->socket = -1;
+		freeaddrinfo(ai);
+		exit(EXIT_FAILURE);
+	}
+
+	if(serve->socket_family == AF_UNSPEC)
+		serve->socket_family = ai->ai_family;
 
-	af = AF_INET;
 #ifdef WITH_SDP
 	if ((serve->flags) && F_SDP) {
-		af = AF_INET_SDP;
+		if (ai->ai_family == AF_INET)
+			ai->ai_family = AF_INET_SDP;
+		else (ai->ai_family == AF_INET6)
+			ai->ai_family = AF_INET6_SDP;
 	}
 #endif
-	if ((serve->socket = socket(af, SOCK_STREAM, IPPROTO_TCP)) < 0)
+	if ((serve->socket = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol)) < 0)
 		err("socket: %m");
 
 	/* lose the pesky "Address already in use" error message */
@@ -1515,20 +1722,15 @@ void setup_serve(SERVER *serve) {
 	}
 
 	DEBUG("Waiting for connections... bind, ");
-	addrin.sin_family = AF_INET;
-#ifdef WITH_SDP
-	if(serve->flags & F_SDP) {
-		addrin.sin_family = AF_INET_SDP;
-	}
-#endif
-	addrin.sin_port = htons(serve->port);
-	if(!inet_aton(serve->listenaddr, &(addrin.sin_addr)))
-		err("could not parse listen address");
-	if (bind(serve->socket, (struct sockaddr *) &addrin, addrinlen) < 0)
+	e = bind(serve->socket, ai->ai_addr, ai->ai_addrlen);
+	if (e != 0 && errno != EADDRINUSE)
 		err("bind: %m");
 	DEBUG("listen, ");
 	if (listen(serve->socket, 1) < 0)
 		err("listen: %m");
+
+	freeaddrinfo (ai);
+
 	sa.sa_handler = sigchld_handler;
 	sigemptyset(&sa.sa_mask);
 	sa.sa_flags = SA_RESTART;
@@ -1610,26 +1812,25 @@ void serve_err(SERVER* serve, const char* msg) {
 void dousers(void) {
 	struct passwd *pw;
 	struct group *gr;
+	gchar* str;
 	if(rungroup) {
 		gr=getgrnam(rungroup);
 		if(!gr) {
-			g_message("Invalid group name: %s", rungroup);
-			exit(EXIT_FAILURE);
+			str = g_strdup_printf("Invalid group name: %s", rungroup);
+			err(str);
 		}
 		if(setgid(gr->gr_gid)<0) {
-			g_message("Could not set GID: %s", strerror(errno));
-			exit(EXIT_FAILURE);
+			err("Could not set GID: %m"); 
 		}
 	}
 	if(runuser) {
 		pw=getpwnam(runuser);
 		if(!pw) {
-			g_message("Invalid user name: %s", runuser);
-			exit(EXIT_FAILURE);
+			str = g_strdup_printf("Invalid user name: %s", runuser);
+			err(str);
 		}
 		if(setuid(pw->pw_uid)<0) {
-			g_message("Could not set UID: %s", strerror(errno));
-			exit(EXIT_FAILURE);
+			err("Could not set UID: %m");
 		}
 	}
 }
@@ -1658,6 +1859,8 @@ void glib_message_syslog_redirect(const gchar *log_domain,
         break;
       case G_LOG_LEVEL_DEBUG:
         level=LOG_DEBUG;
+      default:
+        level=LOG_ERR;
     }
     syslog(level, message);
 }
@@ -1684,7 +1887,9 @@ int main(int argc, char *argv[]) {
 	servers = parse_cfile(config_file_pos, &err);
 	
 	if(serve) {
-		g_array_append_val(servers, *serve);
+		serve->socket_family = AF_UNSPEC;
+
+		append_serve(serve, servers);
      
 		if (!(serve->port)) {
 			CLIENT *client;
